@@ -40,18 +40,18 @@ class AverageMeter:
 
 class Logger():
     loggers = {}
-    def __init__(self, use_wandb=True, wandb_entity=None, wandb_project=None, wandb_name=None, wandb_watch=False
-                     , args=None, model=False
-                     , save=True):
-        self.use_wandb = use_wandb
-        if use_wandb:
+    def __init__(self, use_wandb=True, wandb_entity=None, wandb_project=None, wandb_name=None
+                 , wandb_watch=False, main_process=False
+                 , args=None, model=False
+                 , save=True):
+        self.use_wandb = use_wandb and main_process
+
+        if self.use_wandb:
             wandb.init(entity=wandb_entity, project=wandb_project, name=wandb_name, resume="allow",
-                       config=args, id=wandb_project+"+"+wandb_name)
+                       config=args)
             if wandb_watch and model: wandb.watch(model, log='all')
             if save and args: torch.save({'args': args, }, os.path.join(wandb.run.dir, "args.pt"))
 
-
-
     def getLog(self, k, return_type =None):
         if return_type == "avg":
             return self.loggers[k].avg
@@ -67,8 +67,7 @@ class Logger():
     def resetLog(self):
         self.loggers = {k:AverageMeter() if type(v) == AverageMeter else v for k, v in self.loggers.items()}
 
-
-    def addLog(self, datas:dict, epoch=None, bs = 1):
+    def addLog(self, datas:dict, epoch=None, mae_task_type = None, bs = 1):
         for k, v in datas.items():
             data_type = v[0]
             if data_type == 0:  # Values
@@ -80,9 +79,9 @@ class Logger():
             elif data_type == 2: # Table
                 columns = list(v[1].keys())
                 data_num = len(list(v[1].values())[0])
-                self.loggers[k] = wandb.Table(columns=["epoch"] + columns)
+                self.loggers[k] = wandb.Table(columns=["epoch", "MTT"] + columns)
                 for idx in range(data_num):
-                    self.loggers[k].add_data(str(epoch), *[wandb.Image(to_np(data2PIL(v[1][k][idx]))) if len(v[1][k].shape) == 4 else to_np(v[1][k])[idx] for k in columns])
+                    self.loggers[k].add_data(str(epoch), str(mae_task_type), *[wandb.Image(to_np(data2PIL(v[1][k][idx]))) if len(v[1][k].shape) == 4 else to_np(v[1][k])[idx] for k in columns])
         return True
 
     def getPath(self):
@@ -91,6 +90,7 @@ class Logger():
     def logWandb(self):
         if self.use_wandb:
             wandb.log({k:v.avg if type(v) == AverageMeter else v for k, v in self.loggers.items()})
+            self.resetLog()
         else:
             print("Wandb is not Working Now")
 
